#ifndef SA_H
#define SA_H

#include "misc.h"
#include "la.h"
#include "x86.h"

// relates to representation of disp or imm part of address mode
enum SA_addresskinds { SA_absent, SA_numeric, SA_alphameric };

struct SA_addresses {
  SA_addresskinds kind;
  int number;     // value if known
  ASM_alfa name;  // character representation
  bool OFFSET; // whether offset operator was applied
  DATA_SIZE dsize; // operand size 
};

struct SA_unpackedlines {
  // source text, unpacked into fields
  bool labelled, errors;
  ASM_alfa labfield;
  int tmplate;
  int mod, reg, rm;
  int SIB;
  SA_addresses address1, address2, address3;
  ASM_strings comment;
};

class SA {
  public:
    void parse(SA_unpackedlines &srcline);
    // Analyzes the next source line into constituent fields

    SA(LA * L);
    // Associates syntax analyzer with its lexical analyzer L

  private:
    LA *Lex;
	X86 x86;
    LA_symbols SYM;
	bool match_all_operands(SA_unpackedlines &srcline, const X86_instr& tmplate );
	// Try to unpack the operands of line into srcline
	// for a template instruction given in tmplate
	// returns true in case of success
bool match_operand(SA_unpackedlines &srcline, SA_addresses& addr, ADR_MODE am);

    void getaddress(SA_addresses &address, bool &errors);
    void GetSym(bool &errors);
};

#endif /*SA_H*/
