// Jednoprolazni asembler za mini86 masinu (podskup 80386 asemblera u zasticenom flat modu)
// D. Bojic 2010
// delovi koda (c) P.D. Terry, Rhodes University, 1996

#include "as.h"



void backpatch(unsigned char mem[], int location, int patch)
{ unsigned int value0, value1, value2, value3;
  value0 = mem[location]; value1 = mem[location+1];
  value2 = mem[location+2]; value3 = mem[location+3];
  int value = value0 | (value1<<8) | (value2<<16) | (value3<<24);
  int newval = value + patch;
  mem[location] = newval & 0xFF;
  mem[location+1] = (newval>>8) & 0xFF;
  mem[location+2] = (newval>>16) & 0xFF;
  mem[location+3] = (newval>>24) & 0xFF;
}

void AS::evaluate(ADR_MODE am, SA_addresses address, int &size, int &value, bool &error)
// determine value of address, possibly from symbol table
{ 
  switch (address.kind)
  { case SA_absent:
      break;
    case SA_numeric:
		size= (am==AM_rm8)?32:address.dsize; // in rm8 mode, address is of disp32 kind
      value = address.number /*% 256*/; break;
    case SA_alphameric:
      DATA_SIZE dummy;
      size=(am==AM_rm8)?32:address.dsize; // in rm8 mode, address is of disp32 kind
	  // todo check if symbol size (dummy) bigger than address.size
		Table->valueofsymbol(address.name, location, value, dummy, error); break;
  }
}

void AS::listcode(void)
// List generated code bytes on source listing
{ int blanks=0; // how much blanks after code bytes for justification
  Srce->writeadr(objline.location, 5);
  if (objline.sprefix >= 0 )
    Srce->writehex(objline.sprefix, 3);
  else
    blanks += 3;
  if (objline.opcode >= AS_err && objline.opcode <= AS_dc)
    blanks += 3;
  else
    Srce->writehex(objline.opcode, 3);
  if (objline.opcode2 >= 0 )
    Srce->writehex(objline.opcode2, 3);
  else
    blanks +=3;
  if (objline.modrm >= 0 )
    Srce->writehex(objline.modrm, 3);
  else
    blanks +=3;
  if (objline.sib >= 0 )
    Srce->writehex(objline.sib, 3);
  else
    blanks +=3;
  if ( objline.opcode == AS_ds )
    Srce->writetext("??", 12);
  else
  {
      blanks +=12;  
	  if (objline.immoffs_size > 0 ) 
	  {
		  Srce->writehex(objline.immoffs & 0xFF , 3);
		  blanks-=3;
	  }
	  if (objline.immoffs_size > 1 )
	  {
		  Srce->writehex((objline.immoffs>>8) & 0xFF , 3);
		  blanks-=3;
	  }
	  if (objline.immoffs_size > 2 )
	  {
		  Srce->writehex((objline.immoffs>>16) & 0xFF , 3);
		  Srce->writehex((objline.immoffs>>24) & 0xFF , 3);
		  blanks-=6;
	  }
  }
  fprintf(Srce->lst, "%*c", blanks, ' ');
}

void AS::listsourceline(SA_unpackedlines &srcline, bool coderequired)
// List srcline, with option of listing generated code
{ if (srcline.errors)
    fprintf(Srce->lst, " Cannot assemble the following line:\n");
  if (coderequired) listcode();
  Srce->writetext(srcline.labfield, 9);
  Srce->writetext(X86::table[srcline.tmplate].mnemonic, 5);
  Srce->writetext(srcline.address1.name, 9);
  Srce->writetext(srcline.address2.name, 9);
  Srce->writetext(srcline.address3.name, 9);
  fprintf(Srce->lst, "%s\n", srcline.comment);
}


void AS::assembleline(SA_unpackedlines &srcline, bool &failure)
// assemble single srcline, checking for possible failure
{ bool ignoreerror;
  objline.location = location;
	objline.sprefix = (srcline.address1.dsize == DS_WORD)? 0x66 : -1;
	
	if ( X86::table[srcline.tmplate].modrm >= 0 && X86::table[srcline.tmplate].modrm <= 7 )
		srcline.reg = X86::table[srcline.tmplate].modrm; // reg field will contain opcode extension

	if ( X86::table[srcline.tmplate].modrm == -1 )
	    objline.modrm  = -1;
	else 
	{
		objline.modrm = (srcline.mod<<6) | (srcline.reg << 3) | srcline.rm;
	}
	objline.sib = srcline.SIB;
    objline.immoffs_size = 0; // will be evaluated in the second pass

	objline.opcode = X86::table[srcline.tmplate].opcode;
	if ( X86::table[srcline.tmplate].plusr )
		objline.opcode += srcline.reg;
	objline.opcode2 = X86::table[srcline.tmplate].opcode2;

  srcline.errors = (srcline.errors || X86::table[srcline.tmplate].opcode == AS_err);
  if (objline.opcode >= AS_err && objline.opcode <= AS_dc) // directives
  { 
	  int dummy; // used when dsize in evaluate() not needed 
	  switch (objline.opcode)
      { case AS_beg:
          location = 0;
          break;
        case AS_org:
		  evaluate(X86::table[srcline.tmplate].am1, srcline.address1, dummy, location, srcline.errors);
          objline.location = location;
          break;
        case AS_ds:
		  if (srcline.labelled) Table->enter(srcline.labfield, location, DS_BYTE);
		  evaluate(X86::table[srcline.tmplate].am1, srcline.address1, objline.immoffs_size, objline.immoffs, srcline.errors);
          location = (location + objline.immoffs) % MAX_LOC;
		  break;
        case AS_nul:
        case AS_err:
          if (srcline.labelled) Table->enter(srcline.labfield, location, DS_DWORD);
          break;
        case AS_equ:
          evaluate(X86::table[srcline.tmplate].am1, srcline.address1, objline.immoffs_size, objline.immoffs, srcline.errors);
		  Table->enter(srcline.labfield, objline.immoffs, (objline.immoffs_size==1)?DS_BYTE:((objline.immoffs_size==2)?DS_WORD:DS_DWORD));
          break;
        case AS_dc:
          if (srcline.labelled) Table->enter(srcline.labfield, location, DS_BYTE);
          evaluate(X86::table[srcline.tmplate].am1, srcline.address1, objline.immoffs_size, objline.immoffs, ignoreerror);          
          mem[location]= objline.immoffs;
          location = (location + 1) % MAX_LOC;
          break;
        case AS_end:
          assembling = false;
          break;
        default:   // IF and MAC are not handled
          failure = true;
          break;
      }
  }
  else // Machine instructions
  { 
	if (srcline.labelled) Table->enter(srcline.labfield, location, DS_DWORD);

	if (objline.sprefix >= 0)
    { mem[location] = objline.sprefix;
	  location = (location + 1) % MAX_LOC;
    }
	mem[location] = objline.opcode;
	location = (location + 1) % MAX_LOC;
    if (objline.opcode2 >= 0)
    { mem[location] = objline.opcode2;
      location = (location + 1) % MAX_LOC;
    }
    if (objline.modrm >= 0)
    { mem[location] = objline.modrm;
      location = (location + 1) % MAX_LOC;
    }
    if (objline.sib >= 0)
    { mem[location] = objline.sib;
      location = (location + 1) % MAX_LOC;
	}

	evaluate(X86::table[srcline.tmplate].am1, srcline.address1, objline.immoffs_size, objline.immoffs, ignoreerror);
	evaluate(X86::table[srcline.tmplate].am2, srcline.address2, objline.immoffs_size, objline.immoffs, ignoreerror);
	evaluate(X86::table[srcline.tmplate].am3, srcline.address3, objline.immoffs_size, objline.immoffs, ignoreerror);

    // correction for immediate operand if size prefix is present
    if ( objline.sprefix == 0x66 && X86::hasImm32Opnd(srcline.tmplate))
        objline.immoffs_size = 2;

    ADR_MODE am1 = X86::table[srcline.tmplate].am1;
	if ( am1 == AM_rel8 || am1 == AM_rel32 )
	{
		int instr_len = X86::table[srcline.tmplate].length + ((objline.sprefix>=0)?1:0);
		int offs = objline.immoffs - objline.location - instr_len;
		if ( !ignoreerror && am1 == AM_rel8 && (offs < -128 || offs > 127) )	
		{   fprintf(Srce->lst, " --- short jump out of range");
		    srcline.errors = true; 
		}
		objline.immoffs = offs; 
	}
    if (objline.immoffs_size > 0)
    { mem[location] = objline.immoffs;
	  location = (location + 1) % MAX_LOC;
    }
    if (objline.immoffs_size > 1)
    { mem[location] = objline.immoffs >> 8;
	  location = (location + 1) % MAX_LOC;
    }
    if (objline.immoffs_size > 2)
    { mem[location] = objline.immoffs >> 16;
	  location = (location + 1) % MAX_LOC;
      mem[location] = objline.immoffs >> 24;
	  location = (location + 1) % MAX_LOC;
    }
  }
  if (srcline.errors) failure = true;
  listsourceline(srcline, true);
}

void AS::firstpass(bool &errors)
// Make first and only pass over source code	
{ SA_unpackedlines srcline;
	location=0;
  assembling = true;
  errors = false;

  while (assembling)
  { Parser->parse(srcline);
    assembleline(srcline, errors);
  }
  Table->printsymboltable(errors);
  if (!errors) 
  	  Table->outstandingreferences(mem, backpatch);
}


void AS::saveobjcode(void)
// Simply print all 256 bytes in 16 rows
{ int nextbyte = 0;
  printf("Generating code ... \n");
  if (*objname == '\0') return;
  FILE *objfile = fopen(objname, "w");
  if (objfile == NULL) objfile = stdout;
  for (;;)
  { for (int j = 1; j <= 16; j++)
    { 
		if ( nextbyte == location )
		{
			putc('\n', objfile);
			goto end;
		}
		fprintf(objfile, " %02X", mem[nextbyte]); 
		nextbyte++; 
    }
    putc('\n', objfile);
  }
end:
  if (objfile != stdout) fclose(objfile);
}


void AS::assemble(bool &errors)
{ printf("Assembling ...\n");
  fprintf(Srce->lst, "(One Pass Assembler)\n\n");
  firstpass(errors);
  saveobjcode();
}

AS::AS(char *sourcename, char *listname, char *version, char* objfilename)
{ 
  Srce    = new SH(sourcename, listname, version);
  Lex     = new LA(Srce);
  Parser  = new SA(Lex);
  Table   = new ST(Srce);
  strcpy(objname, objfilename);

  for (int i = 0; i < MAX_LOC; i++) mem[i] = 0xFF;
  // enter opcodes and mnemonics in ALPHABETIC order
  // done this way for ease of modification later
}
